;;; Animation Extension to Common Music (AECM)
;;; version: public beta 1 - February 24, 2001
;;;
;;; Copyright (c) 2001 Bret Battey -- www.BatHatMedia.com
;;;

;;; **********************
;;; DEFINE CM SYNTAX
;;; **********************

(defsyntax max-script
   :pathname (pathname "test.ms")
   :stream-types `((maxscript-file "ms"))
   :nicknames '(ms))

(defobject maxscript-file (event-file header-mixin)
  ((syntax :initform (find-syntax ':ms)))
  (:parameters start end header))

(defmethod initialize-stream-for-processing :after ((stream maxscript-file))
  (let ((file (slot-value stream 'stream)))
    (format file "-- Common Music output of ~A~%~%" (date-string))
    (when (slot-boundp stream 'header)
      (format file "~A~%" (slot-value stream 'header)))))


;;; **********************
;;; INTERNAL UTILITIES
;;; **********************

(defun ms-at-printer (time stream)
  (let* ((ctime (conv-time time))
	(str (format nil "at time ~A " ctime)))
    (princ str stream)))

(defun conv-time (sval)
  (typecase sval
    (fixnum
     (if (= sval 0) (format t "~%Warning: invalid time -- frame number = 0~%"))
     (format NIL "~Df" sval))
    (single-float 
     (format NIL "~Fs" sval))
    (t sval)))

(defun pos (item list)
  (let ((j 0)
	(found NIL))
    (loop for i in list do
      (if (equal i item)
	  (progn 
	    (setf found T)
	    (loop-finish)))
      (incf j)
      finally (if found (return j) NIL))))

(defun print-rest (l)
  (let* ((rstring
	  (butlast
	   (loop for i in l by #'cddr and
	     j in (cdr l) by #'cddr
	     collect  (format nil "~A" i) into st
	     collect ":" into st
	     collect  (format nil "~A" j) into st
	     collect " " into st
	     finally (return st)))))
    (apply #'concatenate 'string rstring)))


;;; **********************
;;; INITIALIZATION
;;; **********************

(defun init-ms ()
  (setf *ms-object-count* 0)
  (setf *ms-key-names* '())
  (setf *ms-key-count* '())
  )

;;; **********************
;;; OUTPUT
;;; **********************

(defun write-ms (obj-name &optional (file "test.ms"))
  (open-cmd (format nil "~A play NIL" file))
  (mix-cmd (format nil "~A 0.0" obj-name)))

;;; **********************
;;; NODE-CONSTRUCTOR
;;; **********************

(defobject ms-node-constructor (note)
  ((time :initform 0) type name specs)
  (:parameters type name &rest specs))

#|
(setf s1 (object ms-node-constructor type 'sphere name 'sph01 
		specs '(radius 2.0 hemisphere .89)))
; should give an unbound slot error:
(setf s2 (object ms-node-constructor type 'sphere 
		specs '(radius 2.0 hemisphere .89)))
|#

; The 'make' function for each object type sets the start time
; of the object to *ms-object-count*, ensuring that objects
; are written out of the CM container in the order in which
; they were created.

(defun make-node (type name &rest specs)
  (object ms-node-constructor type type name name specs specs 
	  start (incf *ms-object-count*)))

#|
(inspect (make-node 'sphere 'sph01 'radius 2.0 'hemisphere .89))
|#

(defun print-ms-node-constructor (object stream)
  (let* ((type (slot-value object 'type))
	 (name (slot-value object 'name))
	 (specs (slot-value object 'specs)))
    (format stream "~A_~A = ~A name:\"~A\" ~A" type name type name 
	    (print-rest specs))))

#|
(print-ms-node-constructor 
 (make-node 'sphere 'sph01 '(radius 2.0 hemisphere .89)) *standard-output*)
|#

(defmethod getvar ((object ms-node-constructor))
  (let* ((type (slot-value object 'type))
	 (name (slot-value object 'name)))
    (format nil "~A_~A" type name)))

(defmethod print-object ((object ms-node-constructor) stream)
  (format stream "#<")
  (print-ms-node-constructor object stream)
  (format stream ">"))

#|
(let ((x (make-node 'sphere 'sph01 'radius 2.0 'hemisphere .89)))
  (print-object x *standard-output*))
|#

(defmethod write-event ((object ms-node-constructor) (stream maxscript-file))
  (let ((str (slot-value stream 'stream)))
    (print-ms-node-constructor object str)
    (format str "~%")))

#|

(progn
  (merge foo ()
    (make-node 'sphere 'sph01 'radius 2.0 'hemisphere .89))
  (write-ms "foo"))

|#

;;; **********************
;;; NODE-EDITORS
;;; **********************

(defobject ms-node-editor (note)
  ((time :initform 0) name attribute value (mstime :initform NIL))
  (:parameters name attribute value &optional mstime))

#|
(setf e1 (object ms-node-editor name 'sph01 
		 attribute 'radius value 4.0))
(setf e2 (object ms-node-editor name 'sph02 mstime 13 
		 attribute 'radius value 10.4))
|#

(defun edit-node (name &rest specs)
  (loop for i in specs by #'cddr and
    j in (cdr specs) by #'cddr collect
    (object ms-node-editor name name attribute i value j mstime NIL
		    start (incf *ms-object-count*))))

#|
(edit-node 'foo 'blah 3 'blarg 12.2)

(let ((x (thread nil ()
	  (edit-node 'foo 'blah 3 'blarg 12.2))))
  (list-object x))

(let ((x (thread nil ()
	  (edit-node 'sph1 'radius 2.0 'hemisphere .56))))
  (list-object x))

|#

;;; edit-node-at-time is a special form of edit-node that requires a 
;;; time at which the edit will occur. Will be printed in the AT form, 
;;; creating a keyframe at the given time. Note that using this
;;; technique in AECM gives you no way to reference and edit the
;;; given key frame and its attributes, so normally this is useful
;;; only for quick and dirty animation work.

(defun edit-node-at-time (time name &rest specs)
  (loop for i in specs by #'cddr and
    j in (cdr specs) by #'cddr collect
    (object ms-node-editor name name attribute i value j mstime time
	    start (incf *ms-object-count*))))

#|

(edit-node-at-time 25 'foo 'blah 3)

(edit-node-at-time 25 'foo 'blah 3 'blu 5.6)

(let ((x (thread nil ()
	  (edit-node-at-time 25 'foo 'blah 3 'blu 5.6))))
  (list-object x))

; this should generate an warning since frame=0 is invalid:

(let ((x (thread nil ()
	  (edit-node-at-time 0 'foo 'blah 3 ))))
  (list-object x))
|#

(defun print-ms-node-editor (object stream)
  (let* ((mstime (slot-value object 'mstime))
	 (name (slot-value object 'name))
	 (attr (slot-value object 'attribute))
	 (val (slot-value object 'value)))
    (setf name (if (eq (type-of name) 'ms-node-constructor)
		   (getvar name)
		 (format nil "$~A" name)))
    (if mstime
	(ms-at-printer mstime stream))
    (princ (format nil "~A.~A = ~A" name attr val) stream)))

#|

(let ((x (thread nil ()
	  (edit-node-at-time 0.0 'sph1 'radius 2.0 'hemisphere .56))))
  (print-ms-node-editor (nth-object 0 x) *standard-output*)
  (print-ms-node-editor (nth-object 1 x) *standard-output*))

(let ((x (thread nil ()
	   (setf y (make-node 'sphere 'sph03 'radius 3))
	   (edit-node-at-time 12.5 y 'radius 2.0 'hemisphere .56))))
  (print-ms-node-constructor (nth-object 0 x) *standard-output*)
  (print-ms-node-editor (nth-object 1 x) *standard-output*)
  (print-ms-node-editor (nth-object 2 x) *standard-output*))


|#

(defmethod print-object ((object ms-node-editor) stream)
  (format stream "#<") 
  (print-ms-node-editor object stream) 
  (format stream ">"))

#|

(let ((x (thread nil ()
	  (edit-node 'sph1 'radius 2.0 'hemisphere .56)
	  (edit-node-at-time 3 'sph1 'radius 6.0))))
  (print-object (nth-object 0 x) *standard-output*)
  (print-object (nth-object 1 x) *standard-output*)
  (print-object (nth-object 2 x) *standard-output*))
|#

(defmethod write-event ((object ms-node-editor) (stream maxscript-file))
  (let ((stream-val (slot-value stream 'stream)))
    (print-ms-node-editor object stream-val)
    (princ (format nil "~%") stream-val)))

#|
(setf x (merge foo ()
	  (edit-node 'sph1 'radius 2.0 'hemisphere .56)
	  (edit-node-at-time 45 'sph1 'radius 5.6)))

(let ((x (merge foo ()
	   (edit-node 'sph1 'radius 2.0 'hemisphere .56)
	   (edit-node-at-time 45 'sph1 'radius 5.6))))
  (list-object x))

(progn 
  (merge foo ()
    (edit-node 'sph1 'radius 2.0 'hemisphere .56)
    (edit-node-at-time 45 'sph1 'radius 5.6))
  (write-ms "foo"))

|#

;;; **********************
;;; CONTROLLERS and KEYS
;;; **********************

(defun assign-controller (name attribute &optional (type 'bezier_float))
  (object ms-node-editor name name 
	  attribute (format nil "~A.controller" attribute)
	  value (format nil "~A()" type)
	  mstime NIL
	  start (incf *ms-object-count*)))

#|
(assign-controller 'sph1 'rotation 'tcb_rotation)

|#

;;; A kludge to track the key count for each key
(defvar *ms-key-names* '())
(defvar *ms-key-count* '())

(defobject ms-key-constructor (note)
  ((time :initform 0) name attribute mstime))

#|
(inspect (object ms-key-constructor name 'sph02 mstime 13 
		 attribute 'radius))
|#

(defun make-key (time name attribute &rest specs)
  (let* ((node (format nil "~A~A" name attribute))
	 (j (pos node *ms-key-names*))
	 (n (if j (incf (nth j *ms-key-count*))
	      (progn
		(setf *ms-key-names* (append *ms-key-names* (list node)))
		(setf *ms-key-count* (append *ms-key-count* (list 1)))
		1))))
    ;; make key
    (object ms-key-constructor name name attribute attribute mstime time
	    start (incf *ms-object-count*))
    ;; edit key attributes
    (if specs
	(loop for i in specs by #'cddr and
	  j in (cdr specs) by #'cddr do
	  (let ((attr (format nil "~A.keys[~D].~A" attribute n i)))
	    (object ms-node-editor name name 
		    attribute attr 
		    value j  
		    start (incf *ms-object-count*)
		    ))))))


#|
(let ((x (thread nil ()
	  (make-key 36.3 'sph1 'radius 'value 15))))
  (list-object x))

;should generate warning frame = 0
(let ((x (thread nil ()
	  (make-key 0 'sph1 'radius 'value 15))))
  (list-object x))
|#

(defun print-ms-key-constructor (object stream)
  (let* ((mstime (conv-time (slot-value object 'mstime)))
	 (name (slot-value object 'name))
	 (attr (slot-value object 'attribute)))
    (setf name (if (eq (type-of name) 'ms-node-constructor)
		   (getvar name)
		 (format nil "$~A" name)))
    (format stream "addNewKey ~A.~A.controller ~A" name attr mstime)))

#|
(let ((x (thread nil ()
	   (make-key 36.3 'sph1 'radius 'value 15))))
  (print-ms-key-constructor (nth-object 0 x) *standard-output*))
|#

(defmethod print-object ((object ms-key-constructor) stream)
  (format stream "#<") 
  (print-ms-key-constructor object stream) 
  (format stream ">"))

#|
(let ((x (thread nil ()
	   (make-key 36.3 'sph1 'radius 'value 15))))
  (print-object (nth-object 0 x) *standard-output*)
  (print-object (nth-object 1 x) *standard-output*))
|#

(defmethod write-event ((object ms-key-constructor) (stream maxscript-file))
  (let ((stream-val (slot-value stream 'stream)))
    (print-ms-key-constructor object stream-val)
    (princ (format nil "~%") stream-val)))

#|
(progn
  (merge foo ()
    (make-key 36.3 'sph1 'radius 'value 15))
  (write-ms "foo"))
|#


;;; **********************
;;; MODIFIERS
;;; **********************

(defobject ms-modifier-constructor (note)
  ((time :initform 0) name modifier modname))

#|
(inspect (object ms-modifier-constructor 
		 name 'sph02
		 modifier 'noiseModifier
		 modname 'noise))
|#

(defun add-modifier (name modifier modname)
  (object ms-modifier-constructor 
	  name name 
	  modifier modifier
	  modname modname
	  start (incf *ms-object-count*)))

#|
(add-modifier 'sph02 'noiseModifier 'noise)
|#

(defun print-ms-modifier-constructor (object stream)
  (let* ((name (slot-value object 'name))
	 (modifier (slot-value object 'modifier))
	 (modname (slot-value object 'modname)))
    (setf name (if (eq (type-of name) 'ms-node-constructor)
		   (getvar name)
		 (format nil "$~A" name)))
    (format stream "addModifier ~A (~A name:\"~A\")" name modifier modname)))

(defmethod print-object ((object ms-modifier-constructor) stream)
  (format stream "#<") 
  (print-ms-modifier-constructor object stream) 
  (format stream ">"))

#|
(let ((x (thread nil ()
	   (add-modifier 'sph02 'noiseModifier 'noise))))
  (print-object (nth-object 0 x) *standard-output*))

|#

(defmethod write-event ((object ms-modifier-constructor) 
			(stream maxscript-file))
  (let ((stream-val (slot-value stream 'stream)))
    (print-ms-modifier-constructor object stream-val)
    (princ (format nil "~%") stream-val)))

#|
(progn
  (merge foo ()
    (add-modifier 'sph02 'noiseModifier 'noise))
  (write-ms "foo"))
|#

;;; **********************
;;; CUSTOM STRING
;;; **********************

(defobject ms-string (note)
  ((time :initform 0) string))

(defun add-string (string)
  (object ms-string string string   
	  start (incf *ms-object-count*)))

(defun print-ms-string (object stream)
  (let* ((string (slot-value object 'string)))
    (format stream string)))

(defmethod print-object ((object ms-string) stream)
  (format stream "#<") 
  (print-ms-string object stream) 
  (format stream ">"))

(defmethod write-event ((object ms-string) 
			(stream maxscript-file))
  (let ((stream-val (slot-value stream 'stream)))
    (print-ms-string object stream-val)
    (princ (format nil "~%") stream-val)))

#|
(object ms-string string "blah")
(add-string "animate_on(")

(progn
  (merge foo ()
    (add-string "animate_on(")
    (edit-node-at-time 20 'sph01 :hemisphere .25)
    (add-string ") animate_off"))
  (write-ms "foo"))
|#

;;; **********************
;;; TYPE CONVERTER ROUTINES
;;; **********************

(defun point3 (x y z)
  (format nil "[~F,~F,~F]" x y z))

#|
(point3 3 6 0)
|#

(defun color (r g b)
  (format nil "(color ~D ~D ~D)" r g b))

#|
(color 12 45 89)
|#

(defun quat1 (degree axis)
  (let ((axis-string (case axis
		       ('x 'x_axis)
		       ('y 'y_axis)
		       ('z 'z_axis))))
  (format nil "(quat ~F ~A)" degree axis-string)))

#|
(quat1 60 'x)
|#

;;; NOTE: angleaxis must be used rather than quat1 to describe a rotation
;;; greater than 360 degrees.

(defun angleaxis (degree axis)
  (let ((axis-string (case axis
		       ('x 'x_axis)
		       ('y 'y_axis)
		       ('z 'z_axis))))
  (format nil "(angleaxis ~F ~A)" degree axis-string)))

#|
(angleaxis 470 'y)
|#

(defun eulerangles (x y z)
  (format nil "(eulerangles ~F ~F ~F)" x y z))

#|
(eulerangles 12 34 56)
|#


;;; **********************
;;; UTILITIES
;;; **********************


; A utility to help create dynamic node names, typically
; in a loop where nodes will be created such as
; sph01 sph02 sph03 sph04

(defun nodenum (name1 number &optional (name2 NIL))
  (if name2
      (format nil "~A~A~A" name1 number name2) 
    (format nil "~A~A" name1 number)))

#|
(list-object
 (thread nil ()
   (loop for i below 4 do
      (make-node 'sphere (nodenum 'sph i)))))

(list-object
 (thread nil ()
   (loop for i below 4 do
      (edit-node 
       (nodenum 'sph i ".modifiers[#myBend]") 'center (point3 3 4 5)))))

|#


;;; get-available-object
;;;
;;; Use to track transient objects to tell when an object's
;;; visible duration has expired and is available to be
;;; utilized for another instance.
;;; starttime = starttime of desired new object
;;; endtime = endtime of desired new object
;;; datalist = a (global) variable containing a list, initiated
;;;   to an empty list. This list will grow to contain the
;;;   last endtime for all objects
;;; This global variable needs to be set to the first value
;;; returned by get-available-object.
;;; Algorithm ensures that objects are only considered available
;;; for use after one frame's worth of time from its endtime.
;;; Can't mix frames and seconds - must use one or the other.

(defun get-available-object (starttime endtime datalist)
  (let* ((dlen (length datalist))
	 (snum (position 
		(- starttime 
		   ;; Need to ensure that
		   ;; an object will least one frame
		   ;; to change its status. If working in frames,
		   ;; subtract 1 frame from starttime. In seconds,
		   ;; subtract length of one frame.
		   (if (integerp starttime) 1 (/ ms-frames-per-second)))
		   datalist :test #'>)))
    (if snum
	(progn
	  (setf (nth snum datalist) endtime)
	  (values  snum 'NIL datalist))
      (progn
	(setf datalist (append datalist (list endtime)))
	(values dlen 'T datalist)))))

#|

(get-available-object 0 4 '())
=>  0 T (4)

(get-available-object 0 4 '(2))
=>  1 T (2 4)

(get-available-object 3 7 '(2 6))
=>  0 NIL (7 6)

(progn
  (init-ms)
  (merge foo ()
    (let ((ecount 0) (ocount 0)
	  (start 6.0) (dur 4.0) sphname (dlist '()))
      (loop for i below 8 do
	(multiple-value-bind (num neednew newlist)
	    (get-available-object start (+ start dur) dlist)
	  (setf dlist newlist)
	  (format t "~%start ~F end ~F~%" start (+ start dur))
	  (format t "num ~D neednew ~B list ~A~%" num neednew dlist)
	  (setf sphname (nodenum 'sph num))
	  (if neednew
	      (progn
		(incf ocount)
		(make-node 'sphere sphname 'radius 0)
		(assign-controller sphname 'radius 'bezier_float)
		(make-key 0.0 sphname 'radius 
			  'value 0.0 'outtangenttype bz-step)))
	  (make-key start sphname 'radius
		    'value 10 'outtangenttype bz-fast)
	  (make-key (+ start dur) sphname 'radius
		    'value 0 'intangenttype bz-slow 'outtangenttype bz-step))
	(incf ecount)
	(incf start 1)
	(setf dur (+ 2 (mod (+ dur 1) 5))))
  (format t "~%~D events expressed in ~D objects~%~%" ecount ocount)))
  (write-ms "foo" "overlaps.ms"))

|#

;;; **********************
;;; HELPFUL PARAMETERS AND CONSTANTS
;;; **********************

(defconstant *aecm-version* "beta 1 2-24-2001")

(defparameter ms-frames-per-second 30)

(defconstant bz-linear (format nil "\#linear"))
(defconstant bz-smooth (format nil "\#smooth"))
(defconstant bz-step (format nil "\#step"))
(defconstant bz-fast (format nil "\#fast"))
(defconstant bz-slow (format nil "\#slow"))
(defconstant bz-custom (format nil "\#custom"))

(defconstant blur-object (format nil "\#object"))
(defconstant blur-non (format nil "\#none"))
(defconstant blur-image (format nil "\#image"))


;;; **********************
;;; MISC
;;; **********************

#|
Time stamp test

(let ((x (thread nil ()
	   (add-modifier 'sph02 'noiseModifier 'noise)
	   (add-modifier 'sph03 'noiseModifier 'noise)
	   (add-modifier 'sph04 'noiseModifier 'noise)
	   (add-modifier 'sph05 'noiseModifier 'noise)
	   )))
  (tlist-object x))

|#




